import numpy as np
import tqdm as tq

from .sol_ode import sol_ode
from .delay import delay

def PID_controller(model, var0, t, param, ylimit, w, y0 = 0, PID = [0,0,0], Tt = 0, x_index = -1, y_index = -1):

    """
    Create a PID controller to adjust the parameter of input ode function,
    calculate the dataframe using the adjusted parameter.

    Args:
    
        model (callable(y,t,...)): The function computes the derivative of y at t.
        var0 (array): Initial condition on var.
        t (array): A sequence of time points for which to solve for y. The initial
                    value point should be the first element of this sequence.
        param (array, optional): Input parameters to model function           
        ylimit (array, optional): Upper and lower setpoint for control element.
                    
        w (float or int, optional): The w value of the measured element function.      
        y0 (int, optional): Parameter of the PID controller. Defaults to 0.
        PID (array, optional): Parameter of the PID controller. Defaults to 0.
        Tt (float, optional): Specified dead time point.
        
        x_index (int, optional): Index of the measured element in var0.
        y_index (int, optional): Index of control element in param.

    Returns:
        results: The solved variables using adjusted control element.
        y_final: The log of the adjusted control element.
        e: The log of error signal
    """

#    warnings.filterwarnings("ignore")


    # create the y value list with initial value.
    y = y0
    y_final = np.zeros(len(t)-1)
    x = np.zeros(len(t)-1)
    e_final = np.zeros(len(t)-1)
    e = w
    KP, KI, KD = PID

    # empty dataframe for saving results
    results = list(range(len(t)-1))

    #  store the value calculated from last time.
    for i in tq.tqdm(range(len(t)-1),position=0,leave=True):
        
        y_final[i] = y
        param[y_index] = y

        #   get the solved results and get the new kla
        df = sol_ode(model, var0, t[i:i+2], param)

        # save the solved results, drop the duplicated value with the same time index,
        # only keeping the last one.
        
        results[i] = np.transpose(df)[-1]
        var0 = np.transpose(df)[-1]
        x[i] = df[x_index][-1]
        
        if Tt != 0:
            x_delay = delay(t, x, Tt, value = 0)

        else:
            x_delay = x

        e_temp = w - x_delay[i]
        # D component
        D = KD * (e - e_temp) / (t[i+1]-t[i])
        # P component
        e = e_temp
        e_final[i] = e_temp
        P = KP * e
        # I component
        I = KI* e*(t[i+1]-t[1])
        
        y = y0 + P + I + D
        y = max(min(y, ylimit[1]), ylimit[0])
        
    return np.transpose(results), y_final, e


import numpy as np
import tqdm as tq

from .sol_ode import sol_ode
from .delay import delay
from .two_position import two_position

def two_pos_controller(model, var0, t, param, Tt, xlimit, ylimit, x_index = -1, y_index = -1):
    
    """
    Create a two position controller to adjust the parameter of input ode function,
    calculate the dataframe using the adjusted parameter.

    Args:
        model (callable(y,t,...)): The function computes the derivative of y at t.
        time (array): A sequence of time points for which to solve for y . The initial
                      value point should be the first element of this sequence.
        xlimit (list): Setpoints for measured element.
        ylimit (list): Setpoints for control element.
        Tt (float): Specified dead time.
        x_index (int, optional): The column index of measured element in the
                                     solved dataframe. Default to -1.

    Returns:
        results: The solved differential equation using adjusted control element.
        y_final: The log of the adjusted control element.
    """
    
    # create the y value list with initial value.
    y = ylimit[0]
    y_final = np.zeros(len(t))
    x = np.zeros(len(t))

    # empty dataframe for saving results
    results = list(range(len(t)))

    #  store the value calculated from last time.
    for i in tq.tqdm(range(len(t)),position=0,leave=True):
        
        y_final[i] = y
        param[y_index] = y

        #   get the solved results and get the new kla
        df = sol_ode(model, var0, t[i:i+2], param)

        # save the solved results, drop the duplicated value with the same time index,
        # only keeping the last one.
        
        results[i] = np.transpose(df)[-1]
        var0 = np.transpose(df)[-1]
        x[i] = df[x_index][-1]
        
        if Tt != 0:
            x_delay = delay(t, x, Tt, value = 0)

        else:
            x_delay = x

        # control the y value by teo position controller.
        y = two_position(x_delay[i], y, xlimit, ylimit)
        
    return np.transpose(results), y_final


import numpy as np
import matplotlib.pyplot as plt
import scipy as sp
from .sol_ode import sol_ode
from .get_index import get_index

def curve_fit_ode(model,
    var0,
    t,
    xdata,
    ydata,
    param=None,
    param_var0=None,
    guess0=None,
    bounds=None,
    col_index=-1,
):

    """
    Use non-linear least squares to fit results of ordinary differential
    equations (model), to data.

    Args:
        model (callable(y,t,...)): The function computes the derivative of y at t.
        var0 (callable or array): Initial condition of var.
        t (array): A sequence of time points for which to solve for y. The initial
                    value point should be the first element of this sequence.
        xdata (array): The independent variable where the data is measured.
                                Should usually be an M-length sequence or an
                                (k, M)-shaped array for functions with k predictors,
                                but can actually be any object.
        ydata (array): The dependent data, a length M array - nominally
                                f(xdata, ...).
        param (array, optional): Parameters used in the ode model function.
        param_var0 (array, optional): Parameters used in the initial condition function.

        guess0 (array, optional): Initial guess for the parameters (length N). If None,
                                then the initial values will all be 1 (if the number of
                                parameters for the function can be determined using
                                introspection, otherwise a ValueError is raised).
        bounds (2-tuple of array_like, optional): Lower and upper bounds on parameters.
                                                   Defaults to [0.1*guess0, 1.5*guess0].
        col_index (int, optional): The index of fitted results in the solved dataframe
                                    of the model. The default will be the last column of
                                    the dataframe.

    Returns:
        avg: Optimal values for the parameters so that the sum of the squared residuals
                of f(xdata, *popt) - ydata is minimized.
        std: The estimated covariance of avg. The diagonals provide the variance of the
                parameter estimate.
    """
    
    
    if isinstance(ydata, int) | isinstance(ydata, float):
        ydata = np.array([ydata] * len(xdata))
    
    if xdata[0] != 0:
        xdata.loc[-1] = 0
        xdata.index = xdata.index + 1
        xdata = xdata.sort_index()
        ydata.loc[-1] = np.math.nan
        ydata.index = ydata.index + 1
        ydata = ydata.sort_index()
                         
    try:
        df = sol_ode(model, var0(param_var0), t, param)
    except:
        df = sol_ode(model, var0, t, param)
        
    if isinstance(df,list):
        C = df[col_index]
    else:
        C = df[:,col_index]

    plt.figure(100)
    plt.grid()
    plt.plot(xdata, ydata, 'o', markersize = 2)
    plt.plot(t,C)
    
    # set the default bounds
    if (bounds == None) & (guess0 != None):
        bounds = np.empty([2, len(guess0)])
        for i in range(len(guess0)):
            if guess0[i] > 0:
                bounds[0][i] = -3 * guess0[i]
                bounds[1][i] = 3 * guess0[i]
            elif guess0[i] == 0:
                bounds[0][i] = -np.inf
                bounds[1][i] = np.inf
            else:
                bounds[0][i] = 3 * guess0[i]
                bounds[1][i] = -3 * guess0[i]

    # identify the index of initial condition's parameters in input guess0
    try:
        p0 = get_index(param_var0, guess0)
        p0_converse = get_index(guess0, param_var0)
    except:
        pass

    # identify the index of ode function's parameters in input guess0
    try:
        p1 = get_index(param, guess0)
        p1_converse = get_index(guess0, param)
    except:
        pass

    # create the fitted function, changing the parameter to making the ode
    # function solved results fitting the input data
    def func(xdata, *args):

        # change the parameter in the ode function
        param_new = param
        for i in range(len(p1)):
            param_new[p1_converse[i]] = args[p1[i]]

        try:
            # change the parameter in the initial condition
            param_var0_new = param_var0
            for i in range(len(p0)):
                param_var0_new[p0_converse[i]] = args[p0[i]]
            # solve the ode function
            results = sol_ode(model, var0(param_var0_new), xdata, param_new)
        except:
            results = sol_ode(model, var0, xdata, param_new)

        # return the solved result
        try:
                return results[:, col_index][~np.isnan(ydata)]
        except:
                return results[col_index][~np.isnan(ydata)]
        
    avg, std = sp.optimize.curve_fit(func, xdata, ydata[~np.isnan(ydata)], p0=guess0, bounds=bounds)
    try:
         df = sol_ode(model, var0(param_var0), t, param)
    except:
         df = sol_ode(model, var0, t, param)
        
    if isinstance(df,list):
        Cnew = df[col_index]
    else:
        Cnew = df[:,col_index]
        
    plt.figure(100)
    plt.plot(t,Cnew)
    plt.legend(["measured data","model","fitted model"])
    plt.show()
    
    print("average:", avg)
    
    return avg, std


import numpy as np
import tqdm as tq
from .sol_ode import sol_ode
from .len_check import len_check
from .fill_array import fill_array

def MonteCarlo(
    model, var0, t, param=None, param_var0=None, col_index = -1
):

    """
    This function automatically runs your model multiple times while varying
    the specified parameters.

    Args:
        model (callable(y,t,...)): The function computes the derivative of y at t.
        var0 (array): Initial condition on y.
        t (array):  A sequence of time points for which to solve for y. The initial
                    value point should be the first element of this sequence.
        param (array, optional): Parameters used in the ode model function.
        param_var0 (array, optional): Parameters used in the initial condition function.
        
        

    Returns:
        avg, std (float): Mean and standard deviation value of the multiple results.
        results (list): The list contains multiple results running from this function.
    """

    print("Start Monte Carlo simulation...")
    
    # check whether the input values for p_func, p_y0 are used for one parameter or multiple parameters
    if np.any(
        isinstance(param, (list, tuple, np.ndarray))
    ):
        var2 = np.array(param, dtype=object)  # input is a 2d array
    else:
        var2 = np.array([param], dtype=object).reshape(1, -1)  # input is a 1d array

    runs = len_check(var2)
    results = list(range(runs))

    if param_var0 != None:
        if np.any(
            isinstance(param_var0, (list, tuple, np.ndarray))
        ):
            var1 = np.array(param_var0, dtype=object)
        else:
            var1 = np.array([param_var0], dtype=object).reshape(1, -1)

        # get the numbers of running
        runs = max(len_check(var1), len_check(var2))

        # fill the array using the runs number
        var1 = fill_array(var1, runs)

    var2 = fill_array(var2, runs)

    # different tqdm bars

    for i in tq.tqdm(range(runs)):
        var_func = [item[i] for item in var2]
        if param_var0 != None:
            var_y0 = [item[i] for item in var1]
            ydata = sol_ode(model, var0(var_y0), t, var_func)
        else:
            ydata = sol_ode(model, var0, t, var_func)
        data = ydata[col_index]
        results[i] = data[:,-1]
    
    mean = np.mean(results,0)
    stddev = np.mean(results,0)

    return results, mean, stddev


def two_position(x, y, xlimit, ylimit):

    """
    This function compares the value of the measured element with limits and
    generates an adjusted output of the control element. The output value of
    the control element is in the range of setpoints.

    Args:
        x (float, optional): Measured element.
        y (float, optional): Control element.
        xmin (int, optional): Lower setpoint for measured element.
        xmax (int, optional): Upper setpoint for measured element.
        ymin (int, optional): Lower setpoint for control element.
        ymax (int, optional): Upper setpoint for control element.

    Returns:
        y (float, optional): Control elements after adjusted.
    """

    if x > xlimit[-1]:
        y = ylimit[-1]
    elif x < xlimit[0]:
        y = ylimit[0]

    return y


import matplotlib.pyplot as plt
from .sol_ode import sol_ode
from .get_index import get_index


def sensitivity(var, xlabel="time", model=None, var0=None, t =None, param=None, param_var0=None, col_index = -1):
    
    """
    The sensitivity function computes the sensitivity of the results of the model
    to changing parameters. Run the model with all parameters at their specified
    values and get the results R1(t). Adjust parameter Par slightly by adding an
    amount ∆ equal to 0.0001 * Par. Run the model again, calling the results R2(t).
    Compute the sensitivity S(t) by the following formula: S(t) = (R2(t) - R1(t)) / ∆

    Args:
        par (float): Value of parameter.
        xlabel (str): Identify the xlabel of results. There are two options: time or
                        length. The time means the results changing with time. The length
                        means the results changing with the reactor series.
        model (callable(y,t,...)): The function computes the derivative of y at t.
        var0 (array): Initial condition on y.
        t (array): A sequence of time points for which to solve for y. The initial
                    value point should be the first element of this sequence.
        param (array, optional): Parameters used in the ode model function.
        param_var0 (array, optional): Parameters used in the initial condition function.
        col_index (integer, optional): if two variables (e.g. X,S) returns Sensitivity of the last, if not defined differently. 

    Returns:
        array:  Sensitivity of solved results to parameter - par.
    """
    if param_var0 != None:

        # solve the ode function using the initial input parameter
        C = sol_ode(model, var0(param_var0), t, param)

        # times the parameter with 1.0001
        param_var0_new = param_var0
        param_new = param
        if var in param_var0:
            index = get_index(var, param_var0)
            param_var0_new[index[0]] = 1.0001 * var
        if var in param:
            index = get_index(var, param)
            param_new[index[0]] = 1.0001 * var

        # solve the ode function using the changed parameter
        C_sens = sol_ode(model, var0(param_var0_new), t, param_new)
    else:
        # solve the ode function using the initial input parameter
        C = sol_ode(model, var0, t, param)

        # times the parameter with 1.0001
        param_new = param

        if var in param:
            index = get_index(var, param)
            param_new[index[0]] = 1.0001 * var

        # solve the ode function using the changed parameter
        C_sens = sol_ode(model, var0, t, param_new)
    # calculate the sensitivity of the parameter along the time or length
    if xlabel == "time":
        try:
            result = var*(C_sens - C) / (0.0001 * var)
        except:
            result = var*(C_sens[col_index]- C[col_index])/(0.0001 * var)
        
        # plot the results
        plt.figure(10)
        plt.grid()
        plt.xlabel("Time [min]")
        plt.ylabel("Absolute-relative Sensitivity")
        try:
            plt.plot(t,result[:,-1])
        except:
            plt.plot(t,result)
        plt.show()
        
        return result
    elif xlabel == "length":
        
        result = var*(C_sens[-1,:] - C[-1,:]) / (0.0001 * var)
        
        # plot the results
        plt.figure(10)
        plt.grid()
        plt.xlabel("Length [i]")
        plt.ylabel("Absolute-relative Sensitivity")
        plt.plot(range(len(result)),result)
        plt.show()
        return result


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import scipy as sp
import odeintw as ow
import tqdm as tq


def sol_ode(model, var0, t, param):

    """
    Solve a system of ordinary differential equations.
    
    Args:
        model (callable(var,t,param)): The function computes the derivative of y at t.
        var0 (array): Initial condition on var.
        t (array):  A sequence of time points for which to solve for var. The initial
                    value point should be the first element of this sequence.
        param (array): Parameters of the model.

    Returns:
        ar (array): Array containing the value of var for each desired time in t, with the initial value var0 in the first row
    """
    if isinstance(param,list):
        if (len(param) == 1):
            param = param[0]
        else: pass
    else: pass

    if isinstance(var0,list):
        if (len(var0) == 1) :
            var0 = var0[0]
        else: pass
    else: pass
    
    results = ow.odeintw(model, y0=var0, t=t, args=(param,))
    
    out = model(var0, t, param) #this is a workaround, so in the case of 1 variable and n CSTRs the result does not get flipped
    
    if isinstance(out, tuple):
        m = int(np.shape(var0)[0])
        ar = list(range(m))
        for i in range(m):
            ar[i] = results[:,i]
    else:
        ar = results
    return ar


import numpy as np
import scipy as sp

def fft(data):

    """
    Compute the 1-D discrete Fourier Transform. 
    
    Args:
        data (dataframe): The dataframe with two columns, time and value
    Returns:
        frq (array): x-axis of the descrete frequency spectrum 
        amp (array): y-axis of the descrete frequency spectrum 
    """

    # acess data frame
    x = data.iloc[:, 0]
    y = data.iloc[:, 1]
    
    # sampling interval
    si = np.median(np.diff(x))
    
    # sampling frequency
    sf = 1/si
    
    # frequencies (x-axis)
    N = len(y)
    k = np.arange(int(N/2))
    p  = N/sf
    frq = k/p
    
    # amplitude (y-axis)
    amp = sp.fft.fft(np.array(y))
    
    # normalize amplitude
    amp = abs(amp/len(y)*2)
    amp[0] = amp[0]/2
    
    # exclude sampling frequency from amplitude
    amp = amp[range(int(len(y)/2))]
    
    # output
    return frq, amp


import numpy as np

def delay(time, var, delay, value = 0):

    """
    This function returns a time array shifted by the delay value

    Args:
        time (array): A sequence of time points.
        var (array): A sequence of values which are shifted in time.
        delay (float): Specified delay time.
        value (float, optional): specified value for var for the delay time
        

    Returns:
        t_delayed (array): Returns a time vector delayed by the delay time with the same length as the time vector.
    """
    DT = time[1]-time[0]
    i_delay = int(round(delay/DT))
        
    var_delayed = np.zeros(len(var))
    var_delayed[0:i_delay] = value
    var_delayed[i_delay:] = var[0:-i_delay]
    
    return var_delayed


import numpy as np

def nextprime(n):
    """
    target: finds the next prime number
    param n: any number (int)
    return: prime number (int)
    """ 
    p=n+1
    for i in range(2,p):
        if(np.mod(p,i)==0):
            p=p+1
    else:
        return p